package org.kantega.jexmec.osgi;

/*
 * Copyright 2011 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.kantega.jexmec.PluginClassLoaderProvider;
import org.osgi.framework.*;

import java.io.IOException;
import java.net.URL;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;

import static java.util.Collections.singleton;

/**
 * Implementation of {@link PluginClassLoaderProvider} that
 * discovers and removes class loaders from OGSi bundles as they are started and stopped.
 * After this provider is started, {@link #getActivator()}  can be called and added as bundle activator to
 * your OSGi implementation.
 */
public class OsgiPluginClassLoaderProvider<P> implements PluginClassLoaderProvider {

    private BundleActivator activator;
    private ClassLoader parentClassLoader;

    public void start(Registry registry, ClassLoader parentClassLoader) {
        activator = new PluginTrackerActivator(registry);
        this.parentClassLoader = parentClassLoader;
    }

    public void stop() {

    }

    private class PluginTrackerActivator implements BundleActivator, SynchronousBundleListener {

        private final Map<Long, ClassLoader> loaders = new HashMap<Long, ClassLoader>();
        private final Registry registry;

        public PluginTrackerActivator(Registry registry) {
            this.registry = registry;
        }

        public void start(BundleContext bundleContext) throws Exception {
            bundleContext.addBundleListener(this);

        }

        public void stop(BundleContext bundleContext) throws Exception {
            bundleContext.removeBundleListener(this);
        }

        public void bundleChanged(BundleEvent bundleEvent) {
            final int eventType = bundleEvent.getType();
            final long bundleId = bundleEvent.getBundle().getBundleId();

            if(bundleId != 0) {
                if (BundleEvent.STARTED == eventType) {
                    ClassLoader bundleDelegatingClassLoader = createBundleClassLoader(bundleEvent.getBundle());

                    registry.add(singleton(bundleDelegatingClassLoader));
                    loaders.put(bundleId, bundleDelegatingClassLoader);

                } else if (BundleEvent.STOPPING == eventType) {
                    ClassLoader loader = loaders.get(bundleId);
                    registry.remove(singleton(loader));
                    loaders.remove(bundleId);
                }
            }

        }
    }


    public BundleActivator getActivator() {
        return activator;
    }

    private ClassLoader createBundleClassLoader(final Bundle bundle) {
        return new BundleDelegatingClassLoader(bundle, parentClassLoader);
    }

    private static class BundleDelegatingClassLoader extends ClassLoader {
        private final Bundle bundle;

        BundleDelegatingClassLoader(Bundle bundle, ClassLoader parent) {
            super(parent);
            this.bundle = bundle;
        }

        @Override
        protected Class<?> findClass(String name) throws ClassNotFoundException {
            return bundle.loadClass(name);
        }

        @Override
        protected URL findResource(String name) {
            return bundle.getResource(name);
        }

        @Override
        @SuppressWarnings("unchecked")
        protected Enumeration<URL> findResources(String name) throws IOException {
            return bundle.getResources(name);
        }
    }
}
